/*****************************************************************************/
/* Copyright (C) 2003-2012, JOANNEUM RESEARCH Forschungs Ges.m.b.H.          */
/* Steyrergasse 17, A-8010 Graz, Austria.                                    */
/*                                                                           */
/*  This file is part of the JRS MPEG-7 Library.                             */
/*  http://mpeg-7.joanneum.at                                                */
/*                                                                           */
/*  JRS MPEG-7 Library is free software: you can redistribute it and/or      */
/*  modify it under the terms of the GNU Lesser General Public License       */
/*  (LGPL) as published by the Free Software Foundation, either version 3    */
/*  of the License, or (at your option) any later version.                   */
/*                                                                           */
/*  JRS MPEG-7 Library is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             */
/*  GNU Lesser General Public License for more details.                      */
/*                                                                           */
/*  You should have received a copy of the GNU Lesser General Public License */
/*  along with this library. If not, see <http://www.gnu.org/licenses/>.     */
/*****************************************************************************/



#ifndef MP7JRS_CLIENT_MANAGER_H
#define MP7JRS_CLIENT_MANAGER_H

#include "Mp7JrsDefines.h"
#include "Mp7JrsNotificationListener.h"
#include "Mp7JrsPtr.h"
#include "Mp7JrsNode.h"

#include <xercesc/util/XMLString.hpp>
#include "Mp7JrsValueVectorOf.h"

// NOTE:
// In order to deal with xerces-c > 2.2.0
// *DON'T* use <KeyValuePair.hpp>!!!
// Use "Mp7JrsKeyValuePair.h" instead.
// Each stable version above 2.4. declares operator =() in KeyValuePair as private 
// which is consistent in terms of memory management
// but inconsistent in terms of usability.
// i.e.: it makes a KeyValuePair unusable for all xerces collection classes ...
// As it turns out, the utility template KeyValuePair is not used in Xerces itself.
#include "Mp7JrsKeyValuePair.h" 


typedef Mp7JrsKeyValuePair<Mp7JrsClientID,Mp7JrsNotificationListener*> ClientListenerPair;
typedef Mp7JrsValueVectorOf<ClientListenerPair> ClientMap;

typedef Mp7JrsKeyValuePair<Mp7JrsClientID,Mp7JrsNodePtr> ClientNodePair;
typedef Mp7JrsValueVectorOf<ClientNodePair> LockQueue;

typedef Mp7JrsValueVectorOf<Mp7JrsClientID> Mp7JrsClientList;

// BAW 26 02 2004: added explicit == operator for compatibility with VC++ .NET 2003
typedef struct _Subtree {
	Mp7JrsNodePtr node;
	bool recursive;

	bool operator==(const _Subtree& second) const { return node==second.node; }
} Subtree;

// BAW 26 02 2004: derived to add explicit == operator for compatibility with VC++ .NET 2003
class MP7JRS_EXPORT Mp7JrsClientRNodePair : public Mp7JrsKeyValuePair<Mp7JrsClientID,Subtree> {
public:
	Mp7JrsClientRNodePair() : Mp7JrsKeyValuePair<Mp7JrsClientID,Subtree>() {}
	Mp7JrsClientRNodePair(Mp7JrsClientID id, Subtree subtree) : Mp7JrsKeyValuePair<Mp7JrsClientID,Subtree>(id, subtree) {}

	bool operator==(const Mp7JrsClientRNodePair& second) const { return (getKey()==second.getKey()) && (getValue()==second.getValue()) ; }
};

#if defined(XML_VISUALCPP)
// SAT: REMOVEME -- fix this warning when there's more time
#pragma warning( disable : 4275 )
#endif

class MP7JRS_EXPORT Mp7JrsListenerList : public Mp7JrsValueVectorOf<Mp7JrsClientRNodePair> {
public:
	Mp7JrsListenerList();

	Mp7JrsClientList GetClientsToBeNotified(Mp7JrsNodePtr node);

};


/** Manages clients accessing Mp7Jrs and need to register for locks and write access. */
class MP7JRS_EXPORT Mp7JrsClientManager {
friend class Mp7JrsNode;
public:
	/** Get a new client ID.
	  * @param listener Object implementing the listener interface which will be registered with the new ID
	  * to receive notifications.
	  */
	static Mp7JrsClientID GetNewID(Mp7JrsNotificationListener* listener);

	/** Get the listener associated with the given client ID. */
	static Mp7JrsNotificationListener* GetListener(Mp7JrsClientID id);

	static void Unregister(Mp7JrsClientID id);

	// SAT: this needs to be accessible from all node subtypes as well
	static void SendUpdateNotification(Mp7JrsNodePtr sender, Mp7JrsNotificationAction action, Mp7JrsClientID client, Mp7JrsNodePtr second, int index = 0);

 	static void SetNotificationsEnabled(bool enabled);

	static void SetLockingEnabled(bool enabled);

	static bool IsNotificationsEnabled();

	static bool IsLockingEnabled();

protected:

	static void QueueForLock(Mp7JrsNodePtr sender, Mp7JrsClientID client, bool forAdd);

	// send notifications

	static bool SendLockRequestNotification(Mp7JrsNodePtr sender, Mp7JrsClientID client, bool forAdd, bool notify);

	static void SendLockNotification(Mp7JrsNodePtr sender, Mp7JrsClientID client, bool forAdd);

	static void SendUnlockNotification(Mp7JrsNodePtr sender);
	
	// register / unregister on node level
	static void Register(Mp7JrsClientID id, Mp7JrsNotificationAction action, Mp7JrsNodePtr node, bool recursive = false);
	
	static void Unregister(Mp7JrsClientID id, Mp7JrsNotificationAction action, Mp7JrsNodePtr node);
	

private:
	static Mp7JrsClientID	_maxID;
	static ClientMap		_clientMap;	
	
	static LockQueue		_lockQueue;

	static Mp7JrsListenerList		_lockRequested;
	static Mp7JrsListenerList		_lockGranted;
	static Mp7JrsListenerList		_unlocked;
	static Mp7JrsListenerList		_added;
	static Mp7JrsListenerList		_inserted;
	static Mp7JrsListenerList		_replaced;

	static bool				_lockingEnabled;
	static bool				_notificationsEnabled;
};

#endif /* MP7JRS_CLIENT_MANAGER_H */
