/*****************************************************************************/
/* Copyright (C) 2003-2012, JOANNEUM RESEARCH Forschungs Ges.m.b.H.          */
/* Steyrergasse 17, A-8010 Graz, Austria.                                    */
/*                                                                           */
/*  This file is part of the JRS MPEG-7 Library.                             */
/*  http://mpeg-7.joanneum.at                                                */
/*                                                                           */
/*  JRS MPEG-7 Library is free software: you can redistribute it and/or      */
/*  modify it under the terms of the GNU Lesser General Public License       */
/*  (LGPL) as published by the Free Software Foundation, either version 3    */
/*  of the License, or (at your option) any later version.                   */
/*                                                                           */
/*  JRS MPEG-7 Library is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             */
/*  GNU Lesser General Public License for more details.                      */
/*                                                                           */
/*  You should have received a copy of the GNU Lesser General Public License */
/*  along with this library. If not, see <http://www.gnu.org/licenses/>.     */
/*****************************************************************************/


#ifndef MP7JRS_SCHEMALOCATIONMAP_H
#define MP7JRS_SCHEMALOCATIONMAP_H

#include "Mp7JrsDefines.h"
#include <xercesc/util/XMLString.hpp>
#include "Mp7JrsKeyValuePair.h" // To support Xerces 2.2.0 AND higher!
#include "Mp7JrsRefVectorOf.h"

typedef struct _sp {
	_sp()
	{
		location = 0;
		prefix_it = 0;
	}

	Mp7JrsRefVectorOf<XMLCh>* prefixes;
	XMLCh* location;
	int prefix_it;
} SchemaProperties;

typedef Mp7JrsKeyValuePair<XMLCh*,SchemaProperties*> SchemaData;

class MP7JRS_EXPORT Mp7JrsSchemaLocationMap
{

public: 
	/** Constructs a location map */
	Mp7JrsSchemaLocationMap(void);
	~Mp7JrsSchemaLocationMap(void);

	/** Sets location for specified schema. If there is an existing entry for this schema, it will be overwritten. 
	  * Otherwise a new entry will be created. The strings passed as arguments will be copied.
	  * @param namespaceURI Namespace name
	  * @param prefix default namespace prefix (including xmlns:) // TODO make it aware, but remove it
	  * @param schemaLocation URN of the schema file. If this is an empty string, no entry will be added to the
	  * string generated by ToText.
	  */
	void SetSchemaLocation(const XMLCh* namespaceURI, const XMLCh* prefix, const XMLCh* schemaLocation);

	/** Adds a prefix alias for a namespace already defined. If the namespace is not already defined, it
	  * will NOT be created.
	  * @param namespaceURI Name of the schema namespace.
	  * @param prefix The prefix alias for this namespace (including xmlns:) // TODO Make it aware but remove it
	  */
	void AddPrefixAlias(XMLCh* namespaceURI, XMLCh* prefix);

	/** Sets location for specified schema. If there is an existing entry for this schema, it will be overwritten. 
	  * Otherwise a new entry will be created. The strings passed as arguments will be copied.
	  * @param namespaceURI Namespace name
	  * @param prefix default namespace prefix (including xmlns:) // TODO 
	  * @param schemaLocation URN of the schema file. If this is an empty string, no entry will be added to the
	  * string generated by ToText.
	  */
	void SetSchemaLocation(const char* namespaceURI, const char* prefix, const char* schemaLocation);

	/** Adds a prefix alias for a namespace already defined. If the namespace is not already defined, it
	  * will NOT be created.
	  * @param namespaceURI Name of the schema namespace.
	  * @param prefix The prefix alias for this namespace (including xmlns:) // TODO removeit 
	  */
	void AddPrefixAlias(const char* namespaceURI, const char* prefix);

	/** Gets the location of the specified schema.
	  * @param namespaceURI Namespace name
	  * @return The URN of the schema file. Returns NULL if no location for this schema has been found.
	  */
	const XMLCh* GetSchemaLocation(const XMLCh* namespaceURI);

	/** Gets the default namespace prefix of the specified schema.
	  * @param namespaceURI Namespace
	  * @return The default prefix of the namespace. Returns NULL if no prefix for this schema has been found.
	  */
	const XMLCh* GetFirstSchemaPrefix(const XMLCh* namespaceURI);

	/** Gets the next namespace prefix of the specified schema.
	  * @param namespaceURI Namespace
	  * @return The next prefix of the namespace. Returns NULL if no more prefixes for this schema has been found.
	  */
	const XMLCh* GetNextSchemaPrefix(const XMLCh* namespaceURI);

	/** Gets the first namespace's name and prefix.
	  * @return true if values set in the arguments, false if no namespace has been found.
	  */
	bool GetFirstNamespace(const XMLCh* * prefix, const XMLCh* * ns);

	/** Gets the next namespace's name and prefix.
	  * @return true if values set in the arguments, false if no more namespace has been found.
	  */
	bool GetNextNamespace(const XMLCh* * prefix, const XMLCh* * ns);

	/** Returns the content of the schema location map as space separated pairs of namespace and schema locations 
	  * It is the responsibility of the caller to deallocate the string.
	  */
	XMLCh* ToText() const;

	/** Returns the number of schema location map entries. 
	  */
	unsigned int size() const {return _locationMap->size();}


	/** Sets the default namespace. If there is an existing entry it will be overwritten.
	  * @param namespaceURI Namespace name
	  * @param prefix default namespace prefix, if any
	  */
	void SetDefaultNamespace(const char * namespaceURI, const char * prefix);

	/** Sets the default namespace. If there is an existing entry it will be overwritten. 
	  * @param namespaceURI Namespace name
	  * @param prefix default namespace prefix, if any
	  */
	void SetDefaultNamespace(const XMLCh* namespaceURI, const XMLCh* prefix);


	/** Gets the namespace's name and prefix of the default namespace.
	  * @param namespaceURI Namespace name
	  * @param prefix default namespace prefix, if any
	  * @return true if values set in the arguments, false if no namespace has been found. The prefix can be null.
	  */
	bool GetDefaultNamespace(const XMLCh* * namespaceURI, const XMLCh* * prefix);

private:
	Mp7JrsRefVectorOf<SchemaData>*	_locationMap;

	unsigned int _nsIter;
	bool _newNs;
	
	XMLCh * _defaultNS;
	XMLCh * _defaultNSPrefix;

	/** Get key value pair for specified schema */
	SchemaData* _GetSchemaData(const XMLCh* namespaceURI);

	/** Deletes a schema properties entry */
	void _DeleteSchemaProperties(SchemaProperties** sp);

};
#endif /* MP7JRS_SCHEMALOCATIONMAP_H */
